-------------------------------------------------------------------------------
-- Title       : Leading-zeroes detector (LZD)
-- Project     : VHDL Library of Arithmetic Units
-------------------------------------------------------------------------------
-- File        : LeadZeroDet.vhd
-- Author      : Reto Zimmermann  <zimmi@iis.ee.ethz.ch>
-- Company     : Integrated Systems Laboratory, ETH Zurich
-- Date        : 1997/12/29
-------------------------------------------------------------------------------
-- Copyright (c) 1998 Integrated Systems Laboratory, ETH Zurich
-------------------------------------------------------------------------------
-- Description :
-- Detection of leading zeroes. Output vector indicates position of the first
-- '1' (starting at MSB).
-- Example: A = "000101" -> Z = "000100".
-- Use the `Encode' component for encoding the output (-> Z = "010").
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
library arith_lib;
use arith_lib.arith_lib.all;

-------------------------------------------------------------------------------

entity LeadZeroDet is

  generic (width : positive := 8;  	-- word width
	   speed : speedType := fast);  -- performance parameter

  port (A : in std_logic_vector(width-1 downto 0);  -- operand
	Z : out std_logic_vector(width-1 downto 0));  -- LZD output

end LeadZeroDet;

-------------------------------------------------------------------------------

architecture Behavioral of LeadZeroDet is

begin

  -- leading-zeroes detection
  lzd : process (A)
    variable zv : std_logic_vector(width-1 downto 0);
  begin
    zv := (others => '0');
    for i in width-1 downto 0 loop
      if A(i) = '1' then
	zv(i) := '1';
	exit;
      end if;
    end loop;
    Z <= zv;
  end process lzd;

end Behavioral;

-------------------------------------------------------------------------------

architecture Structural of LeadZeroDet is

  signal PI, PO : std_logic_vector(width-1 downto 0);  -- prefix prop. in/out
  signal PIT, POT : std_logic_vector(width-1 downto 0);  -- temp.

begin

  -- calculate prefix propagate in signals
  PI <= not A;
  
  -- reverse bit order of PI
  revPI : for i in width-1 downto 0 generate
    PIT(i) <= PI(width-i-1);
  end generate revPI;

  -- solve reverse prefix problem for leading-zeroes detection
  -- (example: "000101" -> "111100")
  prefix : PrefixAnd
    generic map (width, speed)
    port map (PIT, POT);

  -- reverse bit order of PO
  revPO : for i in width-1 downto 0 generate
    PO(i) <= POT(width-i-1);
  end generate revPO;

  -- code output: only bit indicating position of first '1' is '1'
  -- (example: "000101" -> "000100")
  Z(width-1) <= A(width-1);
  Z(width-2 downto 0) <= PO(width-1 downto 1) and A(width-2 downto 0);

end Structural;

-------------------------------------------------------------------------------
