-------------------------------------------------------------------------------
-- Title       : Parallel-prefix subtractor with carry-in, carry-out, zero flag
-- Project     : VHDL Library of Arithmetic Units
-------------------------------------------------------------------------------
-- File        : SubCZ.vhd
-- Author      : Reto Zimmermann  <zimmi@iis.ee.ethz.ch>
-- Company     : Integrated Systems Laboratory, ETH Zurich
-- Date        : 1998/01/11
-------------------------------------------------------------------------------
-- Copyright (c) 1998 Integrated Systems Laboratory, ETH Zurich
-------------------------------------------------------------------------------
-- Description :
-- Binary subtractor using parallel-prefix carry-lookahead logic with:
--   - carry-in (CI), subtracted
--   - carry-out (CO), '1' if subtraction result is negative
--   - zero flag (only valid for CI = 0)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
library synergy;  use synergy.signed_arith.all;
library arith_lib;
use arith_lib.arith_lib.all;

-------------------------------------------------------------------------------

entity SubCZ is

  generic (width : positive := 8;  	-- word width
	   speed : speedType := fast);  -- performance parameter

  port (A, B : in std_logic_vector(width-1 downto 0);  -- operands
        CI : in std_logic;  		-- carry in (subtracted)
	S : out std_logic_vector(width-1 downto 0);  -- sum
        CO, Z : out std_logic);  	-- carry out, zero flag

end SubCZ;

-------------------------------------------------------------------------------

architecture Behavioral of SubCZ is

  signal Auns, Buns, CIuns, Suns : unsigned(width downto 0);  -- unsigned

begin

  -- type conversion: std_logic_vector -> unsigned
  Auns <= conv_unsigned(A, width+1);
  Buns <= conv_unsigned(B, width+1);
  CIuns <= (0 => CI, others => '0');

  -- subtraction
  Suns <= Auns - Buns - CIuns;

  -- type conversion: unsigned -> std_logic_vector
  S <= std_logic_vector(Suns(width-1 downto 0));
  CO <= Suns(width);
  -- calculate zero flag
  Z <= '1' when Suns = 0 else '0';

end Behavioral;

-------------------------------------------------------------------------------

architecture Structural of SubCZ is 

  signal BI : std_logic_vector(width-1 downto 0);  -- B inverted
  signal CII : std_logic;		-- CI inverted
  signal GI, PI : std_logic_vector(width-1 downto 0);  -- prefix gen./prop. in
  signal GO, PO : std_logic_vector(width-1 downto 0);  -- prefix gen./prop. out
  signal PT : std_logic_vector(width-1 downto 0);  -- adder propagate temp

begin

  -- invert B and CI for subtraction
  BI <= not B;
  CII <= not CI;
  
  -- calculate prefix input generate/propagate signal (0)
  GI(0) <= (A(0) and BI(0)) or (A(0) and CII) or (BI(0) and CII);
  PI(0) <= A(0) xor BI(0);
  -- calculate adder propagate signal (0) (PT = A xor B)
  PT(0) <= PI(0);
  -- calculate prefix input generate/propagate signals (1 to width-1)
  preproc : for i in width-1 downto 1 generate
    GI(i) <= A(i) and BI(i);
    PI(i) <= A(i) xor BI(i);
    -- calculate adder propagate signal (1 to width-1) (PT = A xor B)
    PT(i) <= PI(i);
  end generate preproc;

  -- calculate prefix output generate/propagate signals
  prefix : PrefixAndOr
    generic map (width, speed)
    port map (GI, PI, GO, PO);

  -- calculate sum bits, carry-out bits, and zero flag
  S <= PT xor GO(width-2 downto 0) & CII;
  CO <= not GO(width-1);
  Z <= PO(width-1);

end Structural;

-------------------------------------------------------------------------------
